// ============================================================================
//
// Copyright (c) 2010, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================



#include "common.h"

#include "engine.h"
#include "camera.h"
#include "worm.h"
#include "cell.h"
#include "board.h"
#include "spot.h"
#include "head.h"
#include "body_part.h"

#include <time.h>


// global stuff and variables (FIXME get rid of most of them)
// ----------------------------------------------------------------------------

Board* board = NULL; // the board
Worm* worm = NULL; // the worm
Camera* cam = NULL; // the camera
Spot* spot = NULL; // the spot
Engine* engine = NULL; // the navigation engine

// three points in the plan
GLfloat v0[] = { -10.0, -10.0, 0.0 };
GLfloat v1[] = { 10.0, -10.0, 0.0 };
GLfloat v2[] = { 10.0, 10.0, 0.0 };

// shadow matrices
GLfloat light_0_shadow[4][4];

// the board plane
GLfloat plane[4];

unsigned long start_time = 0;
unsigned long display_time = 0;



// initialisation
void init(void) {
    GLfloat ambient_light[] = { 0.4, 0.4, 0.4, 1.0 };
    GLfloat fog_color[] = { 0.0, 0.0, 0.0 };

    GLfloat light_position[3];

    spot->get_light_position(light_position);

    // light 0 (single spot)
    spot->init_spot();

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glShadeModel(GL_SMOOTH);

    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);

    // initialize shadow
    glEnable(GL_FOG);
    {
        glFogi(GL_FOG_MODE, GL_LINEAR);
        glFogfv(GL_FOG_COLOR, fog_color);
        glFogf(GL_FOG_DENSITY, 0.15);
        glHint(GL_FOG_HINT, GL_DONT_CARE);
        glFogf(GL_FOG_START, 3.0 * board_size);
        glFogf(GL_FOG_END, 14.0 * board_size);
    }

    // build matric used to apply shadow
    find_plane(plane, v0, v1, v2);
    shadow_matrix_calc(light_0_shadow, plane, light_position);
}

// keyboard management callback
void keyboard(unsigned char key, int x, int y)
{

    switch (key) {
    case 27:
        exit(0);
        break;
    case 'S':
    case 's':
        cam->start_stop_rotation();
        break;
    case 'O':
    case 'o':
        engine->enable_disable_engine();
        break;
    case 'h':
    case GLUT_KEY_LEFT :
        engine->set_asked_direction (left);
        break;
    case 'k':
    case GLUT_KEY_RIGHT :
        engine->set_asked_direction (right);
        break;
    case 'u':
    case GLUT_KEY_UP :
        engine->set_asked_direction (up);
        break;
    case 'j':
    case GLUT_KEY_DOWN :
        engine->set_asked_direction (down);
        break;
    }
}

void init_game(void);

void fix_motion_realtime_params()
{
    // display_time is the time required for the last display

    // 1. First, let's make sure we'll have exactly 20 display each seconds

    // If each display takes display_time ms, then there is
    // tot_disp_time_p_sec = (1000 ms - 20 * display_time ms)
    // available for sleeping. So we sleep (tot_disp_time_p_sec / 20) ms
    unsigned long tot_disp_time_p_sec = 1000 - (display_each_second * display_time);
    unsigned long sleep_time = tot_disp_time_p_sec / display_each_second;

    //printf("Sleeping %lu ms  \n", sleep_time);

    if(sleep_time < 1000 / display_each_second){
        usleep(sleep_time * 1000);
    }

    // 2. Then let's compute the worm move coefficient

    // Initially, we want the worm to move 3 case each seconds
    // We have 20 display per seconds
    // 3 cases means 300 offset
    // This means it needs to do 300 / 20 = 15 offset each display

    // BUT : we want him to go twice as fast when it reaches 20 body parts,
    // hence :

    float speed = worm_initial_offset_speed
            + (worm_initial_offset_speed * worm->get_body_part_number() / 20);

    // but max is 2.2 times as fast as initial time
    float max_speed = 2.2 * worm_initial_offset_speed;
    if (speed > max_speed) {
        speed = max_speed;
    }
    worm->fixe_coeff(speed);
}

// motion management
void motion(void)
{
    fix_motion_realtime_params();
    start_time = get_time();

    worm->move();

    // change cell in case the offset has been exceeded
    if (worm->get_offset() >= offset_range)
        if (!worm->move_cell())
            // re-init game in case the worm crashed doing the movement
            init_game();

    // call for redisplay
    glutPostRedisplay();
}

// display management callback
void display(void)
{

    glEnable(GL_LIGHTING);
    // clear current buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glLoadIdentity(); // reset transformations

    // position camera and spot
    spot->proceed_position();
    cam->proceed_position();

    glPushMatrix();

        glTranslatef(-5.0 * board->get_x_size(), -5.0 * board->get_y_size(), 0.0);

        // display board and worm
        board->proceed_display();
        worm->proceed_display();

    glPopMatrix();

    // NOW really do display !
    glutSwapBuffers();

    unsigned long end_time = get_time();
    display_time = end_time - start_time;

    //printf ("display time : %lu\n", display_time);
}

// (re-) initialize game.
void init_game(void)
{

    board->init();
    worm->init();
    engine->init();
    cam->init();

    board->create_random_seed();

    worm->set_direction(engine->get_next());
}

// Used by the glut
void reshape(int width, int height)
{
    cam->reshape(width, height);
}

// analyze program argument
int analyse_string(char* string, int &speed_cam, int &speed_worm)
{
    if (string[0] != '-') // params always start by -
        return false;
    else if (&string[1] == NULL)
        return false;
    else if (&string[2] == NULL)
        return false;
    else
        switch ((int) string[1])
        {

        case 115: // speed of the camera
            speed_cam = ((int) string[2] - 48);

        case 118: // speed of the worm
            speed_worm = ((int) string[2] - 48);

        default:
            break;
        }

    return true;
}

// main function
int main(int argc, char** argv)
{

    // initialize random number generator // FIXME find something better
    srand((unsigned) get_time() + rand());

    board = new Board(board_size, board_size);
    worm = new Worm(board_size / 2, board_size / 2, up);
    cam = new Camera;

    engine = new Engine();
    init_game();
    spot = new Spot;

    // glut is used for window management
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(x_res, y_res);
    glutCreateWindow("Snake - by Jerome Kehrli" );

    // initialisation
    init();

    // register callbacks
    glutKeyboardFunc(keyboard);
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(motion);

    printf("\n\nInitialization is OK, graphical window is launched ...\n\n");
    printf(" s -> start/stop camera rotations\n");
    printf(" o -> on/off auto-pilot engine\n\n");
    printf(" When auto-pilot is turned off, one can use u, h, j, k to control the snake.\n\n");

    // and GO !
    glutMainLoop();

    return 0;
}
